program Snapshotter;

(*	Snapshotter -- A program for saving and restoring desktop icons.
	1995-1998 Quinn "The Eskimo!"
	
	This program is distributed as Freeware.
*)

	uses
		Types,
		Memory,
		Resources,
		Files,
		Dialogs,
		Fonts,
		GestaltEqu, 
		Aliases,
		AppleEvents,
		AERegistry,
		AEObjects,
		AEPackObject,
		Folders,
		TextUtils,
		
		ICDebug;

	const
		badVersionErr = -650;			(* snapshot file version number is not 0 *)
		fileBadErr = -651;				(* preposterous alias size *)
		badFileTypeErr = -652;			(* a non-snapshot file was dragged onto Snapshotter *)
		badConfigErr = -653;			(* system software not cool enough to run my code! *)
		noScriptableFinderErr = -654; 	(* no scriptable finder *)

	const
		rMiscStrings = 128;
		strSnapshotSuffix = 1;
		strFullStop = 2;
		strDesktop = 3;
		rErrorStrings = 129;

	const
		rErrorAlert = 128;

	const
		kMyCreator = 'Snp';
		kMyFileType = 'Snp';

	var
		gQuitNow : Boolean;
		gSystemVersion : longint;

	(* ----- File Data Structures ----- *)
	
	// All data structures that are stored in files are aligned
	// with 68K alignment so that we can read directly into them
	// regardless of the structure alignment we're compiled
	// with.
	
	{$ALIGN MAC68K}

	// SnapHeader is the data structure that appears at the immediate
	// start of the snapshot data file.  Following this there is
	// a sequence of folders, each of which begins with a SnapFolderHeader
	// and contains a sequence of SnapRecords, terminated by a SnapRecord
	// with an empty name.
	//
	// The overall structure looks something like:
	//
	//		SnapHeader
	//			SnapFolderHeader
	//				SnapRecord
	//				SnapRecord
	//				SnapRecord (last one has a null name)
	//			SnapFolderHeader
	//				SnapRecord
	//				SnapRecord (last one has a null name)
	//			SnapFolderHeader
	//				SnapRecord
	//				SnapRecord (last one has a null name)
	//		<end of file terminates SnapFolderHeader sequence>
		
	type
		SnapHeader = 
			record
				version : longint;
			end;
	const
		kSnapHeaderSize = sizeof(SnapHeader);
		kSnapHeaderTrueSize = 4;
		
	// SnapFolderHeader is the data structure that begins each
	// folder in the snapshot data file.  It contains an alias
	// to the folder that this snapshot is for.

	type
		SnapFolderHeader =
			record
				aliasSize : longint;
				(* alias data comes after this *)
			end;
	const
		kSnapFolderHeaderSize = sizeof(SnapFolderHeader);
		kSnapFolderHeaderTrueSize = 4;

	// SnapRecord is the data structure used to hold information
	// about one specific file within a folder.  It contains the name of the
	// file and the location of the file within that folder.

	type
		SnapRecord =
			record
				name : Str63;
				loc : Point;
			end;
	const
		kSnapRecordSize = sizeof(SnapRecord);
		kSnapRecordTrueSize = 64 + 4;

	{$ALIGN RESET}
	
	(* ----- Misc Subroutines ----- *)
	
	function GetIndStr(strListID : integer; index : integer) : Str255;
		(* This routine is a basic wrapper around GetIndString that returns
			the string as function result, which is much more useful than
			as a var parameter.  Note that this routine is different from
			all the others in that the string list is specified implicitly by
			resource ID, rather than explicitly by a Handle.
		*)
		var
			tmpString : Str255;
	begin
		GetIndString(tmpString, strListID, index);
		GetIndStr := tmpString;
	end; (* GetIndStr *)

	function DecStr(aNumber : longint) : Str255;
		(*	This is a simple wrapper around the toolbox NumToString routine. *)
		var
			result : Str255;
	begin
		NumToString(aNumber, result);
		DecStr := result;
	end; (* DecStr *)

	function ErrToStr(err : OSErr) : Str255;
		(*	Convert a likely error code into an error string.
		*)
	begin
		case err of
			badVersionErr: 
				ErrToStr := GetIndStr(rErrorStrings, 2);
			fileBadErr: 
				ErrToStr := GetIndStr(rErrorStrings, 3);
			badFileTypeErr: 
				ErrToStr := GetIndStr(rErrorStrings, 4);
			badConfigErr: 
				ErrToStr := GetIndStr(rErrorStrings, 5);
			noScriptableFinderErr: 
				ErrToStr := GetIndStr(rErrorStrings, 6);
			otherwise
				ErrToStr := GetIndStr(rErrorStrings, 1);
		end; (* case *)
	end; (* ErrToStr *)

	procedure DisplayError(err : OSErr);
		(*	Display an error dialog showing the appropriate
			error message and error code.
		*)
		var
			junk : OSErr;
	begin
		ParamText(ErrToStr(err), DecStr(err), '', '');
		junk := StopAlert(rErrorAlert, nil);
	end; (* DisplayError *)

	procedure DisplayAEError(reply : AppleEvent; errNum : OSErr);
		(*	Display an error that's the result of handling an
			AppleEvent.  Basically we put the error number and
			string into the appropriate parameters of the reply
			event.
		*)
		var
			junk : OSErr;
			errStr : Str255;
	begin
		if errNum <> noErr then begin
			if reply.dataHandle <> nil then begin
				errStr := concat(ErrToStr(errNum), GetIndStr(rMiscStrings, strFullStop));
				junk := AEPutParamPtr(reply, keyErrorString, typeChar, @errStr[1], length(errStr));
				ICAssert(junk = noErr);
			end; (* if *)
			if AEInteractWithUser(kAEDefaultTimeout, nil, nil) = noErr then begin
				DisplayError(errNum);
			end; (* if *)
		end; (* if *)
	end; (* DisplayAEError *)
	
	function FSReadQ(refNum : integer; count : longint; buffPtr : univ Ptr) : OSErr;
		(*	Like FSRead except that count is a value parameter. *)
	begin
		FSReadQ := FSRead(refNum, count, buffPtr);
	end; (* FSReadQ *)
	
	function FSWriteQ(refNum : integer; count : longint; buffPtr : univ Ptr) : OSErr;
		(*	Like FSWrite except that count is a value parameter. *)
	begin
		FSWriteQ := FSWrite(refNum, count, buffPtr);
	end; (* FSWriteQ *)
	
	(* ----- AppleEvent Support ----- *)
	
	function AENullDesc : AEDesc;
		(*	A function that returns the empty AEDesc. *)
		var
			result : AEDesc;
	begin
		result.descriptorType := typeNull;
		result.dataHandle := nil;
		AENullDesc := result;
	end; (* AENullDesc *)
	
	procedure AEDisposeDescQ(var desc : AEDesc);
		(*	A wrapper on AEDisposeDesc that throws away the error result. *)
		var
			junk : OSErr;
	begin
		junk := AEDisposeDesc(desc);
		ICAssert(junk = noErr);
	end; (* AEDisposeDescQ *)

	function TypeToDesc(typeCode : DescType; typeData : DescType; var desc : AEDesc) : OSErr;
		(*	Convert an OSType into the corresponding AEDesc. *)
	begin
		TypeToDesc := AECreateDesc(typeCode, @typeData, sizeof(typeData), desc);
	end; (* TypeToDesc *)

	function AEGotRequiredParams(theAppleEvent : AppleEvent) : OSStatus;
		(* Returns no error if you've extracted all of the required
			parameters out of the AppleEvent.
		*)
		var
			typeCode : DescType;
			actualSize : Size;
			err : OSStatus;
	begin
		err := AEGetAttributePtr(theAppleEvent, keyMissedKeywordAttr, typeWildCard, typeCode, nil, 0, actualSize);
		if err = errAEDescNotFound then begin
			err := noErr;
		end else if err = noErr then begin
			err := errAEEventNotHandled;
		end; (* if *)
		AEGotRequiredParams := err;
	end; (* AEGotRequiredParams *)

	function CreateFinderEvent(classID, eventID : AEEventID; var event : AEDesc) : OSErr;
		(*	Creates an AppleEvent that targets the Finder. *)
		var
			target : AEDesc;
			err : OSErr;
	begin
		target := AENullDesc;
		event := AENullDesc;
		err := TypeToDesc(typeApplSignature, 'MACS', target);
		if err = noErr then begin
			err := AECreateAppleEvent(classID, eventID, target, kAutoGenerateReturnID, kAnyTransactionID, event);
		end; (* if *)
		AEDisposeDescQ(target);
		if err <> noErr then begin
			AEDisposeDescQ(event);
		end; (* if *)
		CreateFinderEvent := err;
	end; (* CreateFinderEvent *)

	function OpenItem(fss : FSSpec) : OSErr;
		(*	Tell the Finder to open a specific folder window. We have
			to do this while restoring a snapshot Finder 7.x (at least,
			I didn't try it under Finder 8.x) won't allow you to set
			icon positions in windows that are not open.
		*)
		var
			obj : AEDesc;
			openEvent : AEDesc;
			reply : AEDesc;
			err : OSErr;
	begin
		obj := AENullDesc;
		openEvent := AENullDesc;

		err := AECreateDesc(typeFSS, @fss, sizeof(fss), obj);

		if err = noErr then begin
			err := CreateFinderEvent(kCoreEventClass, kAEOpen, openEvent);
		end; (* if *)
		if err = noErr then begin
			err := AEPutParamDesc(openEvent, keyDirectObject, obj);
		end; (* if *)

		if err = noErr then begin
			err := AESend(openEvent, reply, kAENoReply, kAENormalPriority, kAEDefaultTimeout, nil, nil);
		end; (* if *)

		AEDisposeDescQ(obj);
		AEDisposeDescQ(openEvent);

		OpenItem := err;
	end; (* OpenItem *)

	function SetItemPosition(fullPath : Str255; position : Point) : OSErr;
		(*	Restore the position of the icon in the Finder whose
			full path name is fullPath.
			
			 Note the use of full path names in this routine is not recommended practice!
		*)
		var
			err : OSErr;
			rootDesc : AEDesc;
			fullPathDesc : AEDesc;
			objParentDesc : AEDesc;
			propertyDescriptor : AEDesc;
			objDesc : AEDesc;
			setDataEvent : AEDesc;
			newData : AEDesc;
			reply : AEDesc;
			procDescData : DescType;
	begin
		
		// Prepare for failure.
		
		rootDesc := AENullDesc;
		fullPathDesc := AENullDesc;
		objParentDesc := AENullDesc;
		propertyDescriptor := AENullDesc;
		objDesc := AENullDesc;
		setDataEvent := AENullDesc;
		newData := AENullDesc;
		
		err := AECreateDesc(typeChar, @fullPath[1], length(fullPath), fullPathDesc);

		// Create objParentDesc to specify the parent object, ie item "xxxx" of application
		if (err = noErr) then begin
			err := CreateObjSpecifier(cObject, rootDesc, formName, fullPathDesc, false, objParentDesc);
		end; (* if *)

		// Create propertyDescriptor to hold the property specifier, ie position
		if (err = noErr) then begin
			procDescData := 'posn';
			err := AECreateDesc(typeType, @procDescData, sizeof(procDescData), propertyDescriptor);
		end; (* if *)
		
		// Create objDesc to reference the data to set, ie position of item "xxx" of application
		if (err = noErr) then begin
			err := CreateObjSpecifier(cProperty, objParentDesc, formPropertyID, propertyDescriptor, false, objDesc);
		end; (* if *)

		// Create on descriptor to hold the new data, ie the new icon position.
		if (err = noErr) then begin
			err := AECreateDesc(typeQDPoint, @position, sizeof(position), newData);
		end; (* if *)

		// Create the actual AppleEvent.
		if err = noErr then begin
			err := CreateFinderEvent(kAECoreSuite, kAESetData, setDataEvent);
		end; (* if *)

		// Fill out the parameters, putting the objDesc into the direct object and the new data
		//  into the data parameter.
		if (err = noErr) then begin
			err := AEPutParamDesc(setDataEvent, keyDirectObject, objDesc);
		end; (* if *)
		if (err = noErr) then begin
			err := AEPutParamDesc(setDataEvent, keyAEData, newData);
		end; (* if *)

		// Send the event.
		if (err = noErr) then begin
			err := AESend(setDataEvent, reply, kAEWaitReply, kAENormalPriority, kAEDefaultTimeout, nil, nil);
		end; (* if *)
	
		// Clean up.

		AEDisposeDescQ(rootDesc);
		AEDisposeDescQ(fullPathDesc);
		AEDisposeDescQ(objParentDesc);
		AEDisposeDescQ(propertyDescriptor);
		AEDisposeDescQ(objDesc);
		AEDisposeDescQ(setDataEvent);
		AEDisposeDescQ(newData);

		SetItemPosition := err;
	end; (* SetItemPosition *)
	
	(* ----- Routines for Creating Snapshots ----- *)

	function GetFolderParent(vRefNum : OSErr; dirID : longint; var folder : FSSpec) : OSErr;
		(*	Given a vRefNum and dirID, return the true FSSpec of the denoted folder. *)
		var
			err : OSErr;
			pb : CInfoPBRec;
	begin
		folder.name := '';
		pb.ioNamePtr := @folder.name;
		pb.ioDrDirID := dirID;
		pb.ioVRefNum := vRefNum;
		pb.ioFDirIndex := -1;
		err := PBGetCatInfoSync(@pb);
		if err = noErr then begin
			folder.vRefNum := pb.ioVRefNum;
			folder.parID := pb.ioDrParID;
		end; (* if *)
		GetFolderParent := err;
	end; (* GetFolderParent *)

	function MakeFileName(name, suffix : Str255) : Str31;
		(*	Given a file name and a suffix, append the suffix to
			the name while limiting the result to 31 characters length.
		*)
	begin
		if (length(name) + length(suffix)) <= 31 then begin
			MakeFileName := concat(name, suffix);
		end else begin
			MakeFileName := concat(copy(name, 1, 31 - length(suffix)), suffix);
		end; (* if *)
	end; (* MakeFileName *)

	function CreateSnapfile(where : FSSpec; name : Str255; var snapFile : FSSpec) : OSErr;
		(*	Given a folder that we're about to snapshot and a proposed file name, 
			create a snapshot file with the proposed name (or something close)
			in the right place.  This is either in the folder, for most folders,
			or on the desktop (for a volume snapshot, or if we can't create
			the file in the folder because it's locked).
			Return the name of the created file in snapFile.
		*)
		var
			err, junk : OSErr;
			ref : integer;
			header : SnapHeader;
			attempt : integer;
			snapshotSuffix : Str255;
	begin
	
		// If we're snapshotting the root of a volume, set up where.vRefNum and where.parID
		// to be the desktop folder.
		
		err := noErr;
		if where.parID = 1 then begin
			err := FindFolder(kOnSystemDisk, kDesktopFolderType, true, where.vRefNum, where.parID);
		end; (* if *)
		
		// Loop trying to create a snapshot file, adjusting the name
		// and location appropriately to handle various errors.
				
		if err = noErr then begin
			snapshotSuffix := GetIndStr(rMiscStrings, strSnapshotSuffix);
			attempt := 0;
			where.name := MakeFileName(name, snapshotSuffix);
			repeat
				err := FSpCreate(where, kMyCreator, kMyFileType, 0);
				if err = dupFNErr then begin
					attempt := attempt + 1;
					where.name := MakeFileName(name, concat(snapshotSuffix, ' ', DecStr(attempt)));
				end else if err = afpAccessDenied then begin
					err := FindFolder(kOnSystemDisk, kDesktopFolderType, true, where.vRefNum, where.parID);
					if err = noErr then begin
						err := FSpCreate(where, kMyCreator, kMyFileType, 0);
					end; (* if *)
				end; (* if *)
			until (err <> dupFNErr) or (attempt > 1000);
		end; (* if *)

		// If we succeed, open the file and write out the header.
		
		if err = noErr then begin
			err := FSpOpenDF(where, fsRdWrPerm, ref);
			if err = noErr then begin
				header.version := 0;
				err := FSWriteQ(ref, kSnapHeaderSize, @header);
				junk := FSClose(ref);
				ICAssert(junk = noErr);
			end; (* if *)
			
			// If this errors, we leak the file.
			
		end; (* if *)
		
		// Finally, return the created file to the client.
		
		if err = noErr then begin
			snapFile := where;
		end; (* if *)
		CreateSnapFile := err;
	end; (* CreateSnapFile *)

	function AddFolderToSnapfile(snapFile : FSSpec; folder : FSSpec; isDesktop : Boolean) : OSErr;
		(*	Add the specified folder to the snapFile.  isDesktop controls
			the fudging we have to do with origins if we're snapshotting
			desktop folders.
		*)

		var
			snapFileRef : integer;

		function WriteFileRecord(name : Str63; loc : Point) : OSErr;
			(*	Write a SnapRecord to the current position in the snapFile. *)
			var
				snapRec : SnapRecord;
		begin
			snapRec.name := name;
			snapRec.loc := loc;
			WriteFileRecord := FSWriteQ(snapFileRef, kSnapRecordSize, @snapRec);
		end; (* WriteFileRecord *)

		var
			err, junk : OSErr;
			folderAlias : AliasHandle;
			folderVRef : integer;
			folderID : longint;
			name : Str63;
			folderHeader : SnapFolderHeader;
			fileIndex : integer;
			pb : CInfoPBRec;
			origin : Point;
			fileLoc : Point;
	begin
		folderAlias := nil;
		snapFileRef := 0;

		// Check that the folder exists and get its origin.  The origin
		// is some weird factor that describes how much the folder has
		// been scrolled around.  We factor this out as we write the
		// icon positions to the file, so that when we restore the
		// window the icons appear at the current origin.

		pb.ioNamePtr := @folder.name;
		pb.ioDirID := folder.parID;
		pb.ioVRefNum := folder.vRefNum;
		pb.ioFDirIndex := 0;
		err := PBGetCatInfoSync(@pb);
		if err = noErr then begin
			
			// First setup origin.  This involves more hackery than you
			// might think (-:
		
			// Desktop folders are always considered to be at origin
			// 0,0.  The value on the disk can be out of whack if someone
			// has AppleShare'd into the folder and scrolled it, so we
			// correct that here.
			
			SetPt(origin, 0, 0);
			if not isDesktop then begin
				if (gSystemVersion < $0800) then begin
					origin := pb.ioDrFndrInfo.frScroll;
				end; (* if *)
			end; (* if *)
			
			// Finder 7.x adds an evil fudge factor to the icon positions,
			// probably something to do with the banner at the top of the
			// window.  We get rid of that.  Finder 8.x doesn't do this silliness.
			
			if (gSystemVersion < $0800) then begin
				origin.v := origin.v - 20;
			end; (* if *)
			
			// Now remember the vRefNum and dirID of the folder itself.
			
			folderVRef := folder.vRefNum;
			folderID := pb.ioDirID;
		end; (* if *)
		
		// Open up snapFile and seek to the end.
		
		if err = noErr then begin
			err := FSpOpenDF(snapFile, fsRdWrPerm, snapFileRef);
		end; (* if *)
		if err = noErr then begin
			err := SetFPos(snapFileRef, fsFromLEOF, 0);
		end; (* if *)
		
		// Write the header for this folder.  This involves making an alias
		// and writing out its contents.
		
		if err = noErr then begin
			err := NewAlias(nil, folder, folderAlias);
		end; (* if *)
		if err = noErr then begin
			folderHeader.aliasSize := GetHandleSize(Handle(folderAlias));
			err := FSWriteQ(snapFileRef, kSnapFolderHeaderSize, @folderHeader);
		end; (* if *)
		if err = noErr then begin
			err := FSWriteQ(snapFileRef, folderHeader.aliasSize, Ptr(folderAlias^));
		end; (* if *)
		
		// Now write a record for each file in the folder.
		
		fileIndex := 1;
		repeat
			pb.ioVRefNum := folderVRef;
			pb.ioDirID := folderID;
			pb.ioNamePtr := @name;
			pb.ioFDirIndex := fileIndex;
			err := PBGetCatInfoSync(@pb);
			if err = noErr then begin
				fileLoc := pb.ioFlFndrInfo.fdLocation;
				SubPt(origin, fileLoc);
				err := WriteFileRecord(name, fileLoc);
			end; (* if *)
			fileIndex := fileIndex + 1;
		until err <> noErr;
		
		// Finish by marking the end of the list of files with a sentinel (a
		// SnapRecord with an empty name).
		
		if err = fnfErr then begin
			pb.ioFlFndrInfo.fdLocation.h := 0;
			pb.ioFlFndrInfo.fdLocation.v := 0;
			err := WriteFileRecord('', pb.ioFlFndrInfo.fdLocation);
		end; (* if *)
		
		// Clean up.
		
		if snapFileRef <> 0 then begin
			junk := FSClose(snapFileRef);
			ICAssert(junk = noErr);
		end; (* if *)
		if folderAlias <> nil then begin
			DisposeHandle(Handle(folderAlias));
			ICAssert(MemError = noErr);
		end; (* if *)
		
		AddFolderToSnapfile := err;
	end; (* AddFolderToSnapfile *)

	function SnapshotFolder(fss : FSSpec) : OSErr;
		(*	Snapshot the folder given in fss.  This involves creating
			a snapshot file and adding snapshot data for the folder to it.
			This routine is called in response to open document
			AppleEvent handler.
		*)
		var
			err, junk : OSErr;
			snapFile : FSSpec;
	begin
		err := CreateSnapfile(fss, fss.name, snapFile);
		if err = noErr then begin
			err := AddFolderToSnapfile(snapFile, fss, false);
			if err <> noErr then begin
				junk := FSpDelete(snapFile);
				ICAssert((junk = noErr) | (junk = fnfErr));
			end; (* if *)
		end; (* if *)
		SnapshotFolder := err;
	end; (* SnapshotFolder *)

	function SnapshotDesktop : OSErr;
		(*	Snapshot the desktop.  This is somewhat more complex than
			the above.  We have to create a snapshot file and then
			add a snapshot of the desktop folder on each volume to it.
		*)
		var
			err, junk : OSErr;
			volumeIndex : integer;
			fss : FSSpec;
			snapFile : FSSpec;
			vpb : ParamBlockRec;
	begin
	
		// First find the desktop folder on the system disk.  This is the
		// main desktop folder, the one which we'll use to derive the
		// location of the snapshot file we're about to create.  Note that
		// we pass a constant string (from a resource) of "Desktop" to
		// CreateSnapFile, which appends " Snapshot" to form the default
		// name "Desktop Snapshot".
		
		err := FindFolder(kOnSystemDisk, kDesktopFolderType, false, fss.vRefNum, fss.parID);
		if err = noErr then begin
			err := GetFolderParent(fss.vRefNum, fss.parID, fss);
		end; (* if *)
		if err = noErr then begin
			err := CreateSnapfile(fss, GetIndStr(rMiscStrings, strDesktop), snapFile);
		end; (* if *)
		
		// Now index through each volume on the system.  If the volume
		// has a desktop folder, add it to the snapshot.
		
		if err = noErr then begin
			volumeIndex := 1;
			repeat
				vpb.ioNamePtr := nil;
				vpb.ioVolIndex := volumeIndex;
				err := PBGetVInfoSync(@vpb);
				volumeIndex := volumeIndex + 1;
				if err = noErr then begin
					err := FindFolder(vpb.ioVRefNum, kDesktopFolderType, false, fss.vRefNum, fss.parID);
					if err = noErr then begin
						err := GetFolderParent(fss.vRefNum, fss.parID, fss);
						if err = noErr then begin
							err := AddFolderToSnapfile(snapFile, fss, true);
						end; (* if *)
					end else begin
						err := noErr;
					end; (* if *)
				end; (* if *)
			until err <> noErr;
			
			// If we errored out with nsvErr (no such volume), we just hit the end
			// of the volume list and we swallow the error.
			
			if err = nsvErr then begin
				err := noErr;
			end; (* if *)
			
			// Otherwise, clean up.
			
			if err <> noErr then begin
				junk := FSpDelete(snapFile);
				ICAssert((junk = noErr) | (junk = fnfErr));
			end; (* if *)
		end; (* if *)
		
		SnapshotDesktop := err;
	end; (* SnapshotDesktop *)
	
	(* ----- Routines for Restoring Snapshots ----- *)

	function FSpGetCatInfo (var fss: FSSpec; ioFDirIndex: integer; var cpb: CInfoPBRec): OSStatus;
	begin
		cpb.ioVRefNum := fss.vRefNum;
		cpb.ioDirID := fss.parID;
		cpb.ioNamePtr := @fss.name;
		cpb.ioFDirIndex := ioFDirIndex;
		FSpGetCatInfo := PBGetCatInfoSync(@cpb);
	end; (* FSpGetCatInfo *)

	function FSSPecToFullPath (fss: FSSpec; var path: Str255): OSErr;
		(* Returns a full path for the given FSSpec.  Actually it's an approximation
			(it doesn't handle paths longer than 256 characters) and that's a known
			limitation of the program.
		*)
		var
			err: OSErr;
			pb: CInfoPBRec;
	begin
		err := noErr;
		if fss.parID = 1 then begin
			(* It's a volume, just return the name:. *)
			path := concat(fss.name, ':');
		end else begin
		
			(* It's a file or folder, start by putting the name at the end of the path
				and then iterate up the directory hierarchy adding directory names to
				the front of the path.
			*)
			path := concat(fss.name, ':');
			while (err = noErr) & (fss.parID <> 1) do begin
				err := FSpGetCatInfo(fss, -1, pb);
				path := concat(fss.name, ':', path);
				fss.parID := pb.ioFlParID;
			end; (* while *)
			
		end; (* if *)
		FSSPecToFullPath := err;
	end; (* FSSPecToFullPath *)

	function RestoreSnapshot(snapFile : FSSpec) : OSErr;
		(*	Given a snapshot file, restore the positions of all the icons
			in all the folders contained therein.  In general, we ignore
			errors here, doing the best effort thing.
		*)

		var
			snapFileRef : integer;

		function RestoreFolder(folder : FSSpec) : OSErr;
			(*	This routine reads SnapRecords out of the snapFileRef
				until it hits the terminator.  For each SnapRecord,
				it sends an AppleEvent to the Finder to reposition
				the corresponding icon.
			*)
			var
				err : OSErr;
				done : Boolean;
				snapRec : SnapRecord;
				folderPath : Str255;
		begin
			// Get the full path to the folder.
			
			err := FSSPecToFullPath(folder, folderPath);

			// Tell the Finder to open that window.  This is necessary,
			// at least under Finder 7, because the Finder will not reposition
			// icons properly unless the window is open )-:
			
			if err = noErr then begin
				err := OpenItem(folder);
			end; (* if*)
			
			// Read SnapRecords from the file until we hit the sentinel
			// (ie the name is empty).  For each SnapRecord, restore
			// the position of the icon.
			
			if err = noErr then begin
				done := false;
				repeat
					err := FSReadQ(snapFileRef, kSnapRecordSize, @snapRec);
					if err = noErr then begin
						done := (snapRec.name = '');
						if not done then begin
							err := SetItemPosition(concat(folderPath, snapRec.name), snapRec.loc);
						end; (* if *)
					end; (* if *)
				until done;
			end; (* if *)
			
			RestoreFolder := err;
		end; (* RestoreFolder *)

		var
			err, junk : OSErr;
			aliasH : AliasHandle;
			header : SnapHeader;
			folderHeader : SnapFolderHeader;
			folder : FSSpec;
			junkBool : Boolean;
			tmpstr  : Str255;
	begin
	
		// Prepare for failure.
		
		snapFileRef := 0;
		aliasH := nil;
		
		// Open up the snapshot file.
		
		err := FSpOpenDF(snapFile, fsRdPerm, snapFileRef);

		// Read in and check header.
		
		if err = noErr then begin
			err := FSReadQ(snapFileRef, kSnapHeaderSize, @header);
		end; (* if *)
		if (err = noErr) & (header.version <> 0) then begin
			err := badVersionErr;
		end; (* if *)
		
		// Loop through each folder in the file.  We know we're out of folders
		// when we hit the end of the file.  But first, create an AliasHandle
		// to read in to.
		
		if err = noErr then begin
			aliasH := AliasHandle(NewHandle(0));
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			repeat
			
				// Read the SnapFolderHeader in.  This tells us how big the alias
				// data is, so we can then resize the handle appropriately and read
				// in the alias information as well.
				
				err := FSReadQ(snapFileRef, kSnapFolderHeaderSize, @folderHeader);
				if err = noErr then begin
					if (folderHeader.aliasSize < 0) or (folderHeader.aliasSize > 64000) then begin
						err := fileBadErr;
					end else begin
						SetHandleSize(Handle(aliasH), folderHeader.aliasSize);
						err := MemError;
					end; (* if *)
				end; (* if *)
				if err = noErr then begin
					err := FSReadQ(snapFileRef, folderHeader.aliasSize, Ptr(aliasH^));
				end; (* if *)
				
				// We now have the alias to the folder.  Resolve it and call
				// RestoreFolder to restore it.  RestoreFolder will read the appropriate
				// SnapRecords out of the file, so we can continue on with the next
				// SnapFolderHeader.
				
				if err = noErr then begin
					err := ResolveAlias(nil, aliasH, folder, junkBool);
					if false then begin
					
						// A really evil hack to restore my icons after a hard disk swap.
						// I guess one day I should let everyone benefit from this sort
						// of cleverness.
						
						if err <> noErr then begin
							err := GetAliasInfo(aliasH, asiAliasName, tmpstr);
							if err = noErr then begin
								err := FSMakeFSSpec(0, 0, 'YipYip:Desktop Folder', folder);
							end; (* if *)
						end; (* if *)
					end; (* if *)
				end; (* if *)
				if err = noErr then begin
					err := RestoreFolder(folder);
				end; (* if *)
			until err <> noErr;
		end; (* if *)
		
		// If we ran of the end of the file, that's expected.  OK, so it's a lame
		// file format!
		
		if err = eofErr then begin
			err := noErr;
		end; (* if *)

		// Clean up.
		if snapFileRef <> 0 then begin
			junk := FSClose(snapFileRef);
			ICAssert(junk = noErr);
		end; (* if *)
		if aliasH <> nil then begin
			DisposeHandle(Handle(aliasH));
			ICAssert(MemError = noErr);
		end; (* if *)
		
		RestoreSnapshot := err;
	end; (* RestoreSnapshot *)

	function ProcessOpenedFSSpec(fss : FSSpec) : OSErr;
		(*	A file has been dropped on to our application.  Do the right thing!
			ie If it's a folder, take a snapshot, if it's a snapshot file, restore
			it.
		*)
		var
			err : OSErr;
			cpb : CInfoPBRec;
	begin
		ICDebugStr('ProcessOpenedFSSpec');
		err := FSpGetCatInfo(fss, 0, cpb);
		if err = noErr then begin
			if btst(cpb.ioFlAttrib, 4) then begin
				err := SnapshotFolder(fss);
			end else begin
				if cpb.ioFlFndrInfo.fdType = kMyFileType then begin
					err := RestoreSnapshot(fss);
				end else begin
					err := badFileTypeErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		ProcessOpenedFSSpec := err;
	end; (* ProcessOpenedFSSpec *)
	
	(* ----- Vanilla Required AppleEvent Support ----- *)

	function HandleOAPP(var theAppleEvent, reply : AppleEvent; junk : longint) : OSErr;
		(*	We've been launched without any documents, so snapshot the desktop. *)
		var
			err : OSErr;
	begin
		{$unused junk}

		// We don't expect any params at all, but check in case the client requires any.
		
		err := AEGotRequiredParams(theAppleEvent);
		if err = noErr then begin
			err := SnapshotDesktop;
		end; (* if *)
		gQuitNow := true;
		DisplayAEError(reply, err);
		HandleOAPP := err;
	end; (* HandleOAPP *)

	function HandleDocs(var theAppleEvent, reply : AppleEvent; junk : longint) : OSErr;
		(*	We've been launched with some items (either documents or folders).
			Process each one in turn.
		*)
		var
			myFSS : FSSpec;
			docList : AEDescList;
			index, itemCount : longint;
			actualSize : Size;
			keywd : AEKeyword;
			typeCode : DescType;
			err : OSErr;
	begin
		{$unused junk}

		err := AEGetParamDesc(theAppleEvent, keyDirectObject, typeAEList, docList);
		if err = noErr then begin
			err := AEGotRequiredParams(theAppleEvent);
		end; (* if *)
		if err = noErr then begin
			err := AECountItems(docList, itemCount);
			for index := 1 to itemCount do begin
				if err = noErr then begin
					err := AEGetNthPtr(docList, index, typeFSS, keywd, typeCode, @myFSS, sizeof(myFSS), actualSize);
					if err = noErr then begin
						err := ProcessOpenedFSSpec(myFSS);
					end; (*if *)
				end; (* if *)
			end; (* for *)
			AEDisposeDescQ(docList);
		end; (* if *)
		gQuitNow := true;
		DisplayAEError(reply, err);
		HandleDocs := err;
	end; (* HandleDocs *)

	function HandleQUIT(var theAppleEvent, reply : AppleEvent; junk : longint) : OSErr;
		var
			err : OSErr;
	begin
		{$unused junk}

		err := AEGotRequiredParams(theAppleEvent);
		if err = noErr then begin
			gQuitNow := true;
		end; (* if *)
		DisplayAEError(reply, err);
		HandleQUIT := err;
	end; (* HandleQuit *)

	var
		i : integer;
		err, junk : OSErr;
		junkBool : Boolean;
		response : longint;
		event : EventRecord;
begin
	// Standard toolbox init.
	
	InitGraf(@qd.thePort);
	InitFonts; { foreground only }
	InitWindows; { foreground only }
	InitMenus; { foreground only }
	TEInit; { prforeground only }
	InitDialogs(nil); { foreground only }
	MaxApplZone;
	for i := 1 to 10 do begin
		MoreMasters;
	end; (* for *)

	// Check that we can run.
	
	err := Gestalt(gestaltAppleEventsAttr, response);
	if (err = noErr) and not btst(response, gestaltAppleEventsPresent) then begin
		err := badConfigErr;
	end; (* if *)
	err := Gestalt(gestaltFSAttr, response);
	if (err = noErr) and not btst(response, gestaltHasFSSpecCalls) then begin
		err := badConfigErr;
	end; (* if *)
	err := Gestalt('fndr', response);
	if (err = noErr) and not btst(response, 3) then begin
		err := noScriptableFinderErr;
	end; (* if *)
	if err = noErr then begin
		err := Gestalt(gestaltSystemVersion, gSystemVersion);
	end; (* if *)
	
	// Install AppleEvent handlers.
	
	if err = noErr then begin
		junk := AEInstallEventHandler(kCoreEventClass, kAEOpenApplication, @HandleOAPP, 0, false);
		ICAssert(junk = noErr);
		err := AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, @HandleDocs, 0, false);
		ICAssert(junk = noErr);
		err := AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, @HandleQUIT, 0, false);
		ICAssert(junk = noErr);
	end; (* if *)
	
	// Run main event loop.
	
	if err = noErr then begin
		gQuitNow := false;
		while not gQuitNow do begin
			junkBool := WaitNextEvent(everyEvent, event, maxlongint, nil);
			case event.what of
				kHighLevelEvent: 
					junk := AEProcessAppleEvent(event);
				otherwise
					(* do nothing *) ;
			end; (* case *)
		end; (* while *)
	end else begin
		DisplayError(err);
	end; (* if *)
end. (* Snapshotter *)